package x.awt;

import java.awt.*;
import java.awt.event.*;

/**
 *
 * Custom version of a scrollbar, without beveled edges.
 *
 */
public class FlatScrollbar extends BufferPanel implements Runnable
{
    private double min, max, value;   
    private int thumbY,               // y-position of thumb
                  thumbRadius=17,       // "radius", or half of height of thumb
                  mouseClickOffset;     // offset between thumb's position and last mouse click
    private boolean mouseIsDown;      // whether mouse is down
    private Color lineColor=Color.black;
    private Color bgcolor=Color.lightGray;
    private Color fillColor=Color.gray;
    private Color highlightColor=Color.black;
    private Thread watcher;
    private Point mouse=new Point(-1,-1);
    private double increment;
    private boolean arrowActive=false;
    
    public FlatScrollbar(double min, double max, double increment, double value)
    {
        this.min=min;
        this.max=max;
        this.value=value;
        this.increment=increment;
    }
    
    public void setParams(double min, double max, double increment, double value)
    {
        this.min=min;
        this.max=max;
        this.value=value;
        this.increment=increment;
        findYFromValue();
        redraw();
    }

    public void run()
    {
        for (;;)
        {
            try {Thread.sleep(50);}
            catch (Exception e) {}
            double value=getValue();
            int w=size().width;
            if (mouseIsDown && mouse.x>=0 && mouse.x<w)
            {
                if (mouseInUpArrow())
                    move(-increment);
                else if (mouseInDownArrow())
                    move(increment);
                    
            }
        }
    }
    
    void move(double increment)
    {
        value=Math.min(max, Math.max(min, value+increment));
        int w=size().width, h=size().height;
        findYFromValue();
        redraw();
        fireChangeEvent(false);
    }
    
    private void findYFromValue()
    {
        int w=size().width, h=size().height;
        thumbY=w+thumbRadius+(int)((h-2*(w+thumbRadius))*(value-min)/(max-min));
    }
    
    void fireChangeEvent(boolean takeValueFromThumb)
    {
        if (takeValueFromThumb)
        {
            double w=size().width, h=size().height;
            value=min+(max-min)*(thumbY-thumbRadius-w)/(h-2*(w+thumbRadius));
        }
        postEvent(new Event(this, Event.ACTION_EVENT, new Double(value)));
    }
    
    boolean mouseInUpArrow()
    {
        return mouseIsDown && arrowActive && 
               mouse.x>=0 && mouse.x<=size().width &&
               mouse.y>=0 && mouse.y<=size().width;
    }
    
    boolean mouseInDownArrow()
    {
        int w=size().width, h=size().height;
        return mouseIsDown && arrowActive && 
               mouse.x>=0 && mouse.x<=w &&
               mouse.y>=h-w && mouse.y<h;
    }
    
    /*
    Call to start up watcher thread.
    */
    public void start()
    {
        stop();
        watcher=new Thread(this);
        watcher.start();
    }
    
    /*
    Call to stop the watcher thread.
    */
    public void stop()
    {
        if (watcher!=null) watcher.stop();
    }
    
    /*
    make sure that watcher thread doesn't hang around
    any longer than necessary.
    */
    public void finalize() throws Throwable
    {
        stop();
        super.finalize();
    }
    
    public void reshape(int x, int y, int w, int h)
    {
        thumbY=h/2;
        super.reshape(x,y,w,h);
    }
    
    public void setColors(Color bgcolor, Color lineColor, Color fillColor, Color highlightColor)
    {
        this.bgcolor=bgcolor;
        this.lineColor=lineColor;
        this.fillColor=fillColor;
        this.highlightColor=highlightColor;
        redraw();
    }
    
    public double getValue()
    {
        return value;
    }
    
    /*
    Draw component to offscreen buffer
    */
    protected void drawOffscreen(Graphics g)
    {
        
        int w=size().width, h=size().height;
        g.setColor(bgcolor);
        g.fillRect(0,0,w,h);
        int m=2;
        int a=w/2, b=(2*w)/3, c=w/2-3;
        
        // draw up arrow.
        g.setColor(fillColor);
        g.fillRect(0,0,w,w);
        g.setColor(lineColor);
        g.drawLine(0,w,w,w);
        Polygon up=triangle(a,b-c,a+c,b,a-c,b);
        if (mouseInUpArrow())
            g.fillPolygon(up);
        g.drawPolygon(up);
        
        
        // draw down arrow.
        g.setColor(fillColor);
        g.fillRect(0,h-w,w,w);
        g.setColor(lineColor);
        g.drawLine(0,h-w,w,h-w);
        b=h-(2*w)/3;
        Polygon down=triangle(a,b+c,a+c,b,a-c,b);
        if (mouseInDownArrow())
            g.fillPolygon(down);
        g.drawPolygon(down);
        
        
        g.setColor(fillColor);
        g.fillRect(1,thumbY-thumbRadius,w-2,2*thumbRadius+1);
        g.setColor(lineColor);
        g.drawRect(0,thumbY-thumbRadius,w-1,2*thumbRadius+1);
        /*
        for (int i=-9; i<=9; i+=3)
        {
            int y=thumbY+h/2+i;
            g.setColor(bgcolor);
            g.drawLine(m,y,w-m-1,y);
        }
        */
        g.setColor(lineColor);
        g.drawRect(0,0,w-1,h-1);
        
        Toolkit.getDefaultToolkit().sync();
    }
    
    Polygon triangle(int x1, int y1, int x2, int y2, int x3, int y3)
    {
        int[] xp={x1,x2,x3,x1};
        int[] yp={y1,y2,y3,y1};
        return new Polygon(xp,yp,4);
    }
    
    ////////////////////////////////
    // event handling
    
    public boolean mouseUp(Event e, int x, int y)
    {
        arrowActive=false;
        mouseIsDown=false;
        redraw();
        return true;
    }
    
    public boolean mouseDrag(Event e, int x, int y)
    {
        if (!arrowActive)
        {
            thumbY=clamp(y-mouseClickOffset);
            fireChangeEvent(true);
        }
        mouse.x=x;
        mouse.y=y;
        redraw();
        return true;
    }
    
    
    
    int clamp(int y)
    {
        int w=size().width;
        return Math.max(thumbRadius+w,Math.min(size().height-thumbRadius-w,y));
    }
    
    public boolean mouseDown(Event e, int x, int y)
    {
        mouse.x=x;
        mouse.y=y;
        mouseIsDown=true;
        
        if (mouse.y>size().width && mouse.y<=size().height-size().width)
        {
        
            if (Math.abs(thumbY-y)<thumbRadius)
                mouseClickOffset=y-thumbY;
            else
            {
                thumbY=clamp(y);
                mouseClickOffset=0;
            }
            fireChangeEvent(true);
        }
        else
            arrowActive=true;
        redraw();
        return true;
    }
    
}