package x.awt.image;

import java.awt.*;
import java.awt.image.*;

public class TranslucentGraphics
{
    private static final int OPAQUE=0xff000000;
    private int w,h;
    private short[] pix;
    byte[] imagePix;
    private IndexColorModel model;

    public TranslucentGraphics(int w, int h)
    {
        this(w,h,Color.black,Color.white);
    }

    public TranslucentGraphics(int w, int h, Color background, Color foreground)
    {
        Color[] palette=new Color[256];
        for (int i=0; i<256; i++)
        {
            int r=(foreground.getRed()*i+background.getRed()*(255-i))/255;
            int g=(foreground.getGreen()*i+background.getGreen()*(255-i))/255;
            int b=(foreground.getBlue()*i+background.getBlue()*(255-i))/255;
            palette[i]=new Color(r,g,b);
        }
        init(w,h,palette);
    }

    public TranslucentGraphics(int w, int h, Color[] palette)
    {
        init(w,h,palette);
    }
    
    public byte[] getImagePix()
    {
        return imagePix;
    }
    
    public IndexColorModel getIndexColorModel()
    {
        return model;
    }

    private void init(int w, int h, Color[] palette)
    {
        if (palette.length!=256)
            throw new IllegalArgumentException("TranslucentGraphics must get 256-color palette. Got "+palette.length+" instead!");
        this.w=w;
        this.h=h;
        pix=new short[w*h];
        imagePix=new byte[w*h];
        byte[] r=new byte[256];
        byte[] g=new byte[256];
        byte[] b=new byte[256];
        for (int i=0; i<256; i++)
        {
            r[i]=(byte)palette[i].getRed();
            g[i]=(byte)palette[i].getGreen();
            b[i]=(byte)palette[i].getBlue();
        }
        model=new IndexColorModel(8,256,r,g,b);
    }

    public void fillCircle(int x, int y, int radius, int alpha)
    {
        if (x-radius<0 || y-radius<0 || x+radius>=w || y+radius>=h) return;
        int r2=radius*radius;
        for (int cy=-radius; cy<=radius; cy++)
        {

            int cx=(int)Math.sqrt(r2-cy*cy);
            int b1=(cy+y)*w+x+cx;
            int b2=(cy+y)*w+x-cx;
            pix[b1]-=alpha;
            pix[b2]+=alpha;
        }
    }

    public void fillRect(int x, int y, int width, int height, int alpha)
    {
        if (x<=0 || x+width>=w || y<0 || y+height>=h) return;
            //throw new IllegalArgumentException("rectangle out of bounds.");
        int base=x+y*w;
        for (int i=0; i<height; i++)
        {
            pix[base]+=alpha;
            pix[base+width]-=alpha;
            base+=w;
        }
    }

    public void fillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, int alpha)
    {

        int orientation=(x2-x1)*(y3-y1)-(x3-x1)*(y2-y1)<0 ? 1 : -1;
        edge(x1,y1,x2,y2,alpha*orientation);
        edge(x2,y2,x3,y3,alpha*orientation);
        edge(x3,y3,x1,y1,alpha*orientation);
    }

    private void edge(int x1, int y1, int x2, int y2, int alpha)
    {
        if (y1==y2) return;
        if (y2<y1)
        {
            alpha=-alpha;
            int t=y2;
            y2=y1;
            y1=t;
            t=x2;
            x2=x1;
            x1=t;
        }
        int dy=y2-y1;
        int dx=x2-x1;
        int a=0;
        int base=w*y1+x1;

        if (dx>0)
        {
            for (int i=dy; i>0; i--)
            {
                pix[base]+=alpha;
                a+=dx;
                base+=w;
                while (a>dy)
                {
                    a-=dy;
                    base++;
                }
            }
        }
        else
        {
            for (int i=dy; i>0; i--)
            {
                pix[base]+=alpha;
                a-=dx;
                base+=w;
                while (a>dy)
                {
                    a-=dy;
                    base--;
                }
            }
        }

    }

    public void fillRibbon(boolean up, int x1, int x2, int x3, int x4, int y, int alpha)
    {
        x2=Math.max(x2, x1+2);
        x4=Math.max(x4, x3+2);
        try
        {
            int r1=(x4-x1)/2;
            int r2=(x3-x2)/2;
            halfDisk(up, (x1+x4)/2,y,r1,alpha);
            halfDisk(up, (x2+x3)/2,y,r2,-alpha);
        }
        catch (Exception e) {}
    }


    public void fillFlatRibbon(boolean up, int x1, int x2, int x3, int x4, int y, int alpha)
    {
        x2=Math.max(x2, x1+2);
        x4=Math.max(x4, x3+2);
        int height=Math.min(x3-x1, x4-x2)/2;
        //System.out.println("height="+height);
        //if (height<y) return;
        try
        {
        startHalfEllipse(up, x1,x3,y,height,alpha);
        startHalfEllipse(up, x2,x4,y,height,-alpha);
        }
        catch (Exception e) {}
    }

/*
    public void fillRibbon(boolean up, int x, int y, int inRadius, int outRadius, int alpha)
    {
        if (inRadius>=outRadius)
            throw new IllegalArgumentException("inRadius was >= outRadius.");
        if (x-outRadius<0 || y-outRadius<0 || x+outRadius>=w || y+outRadius>=h) return;
        int d=(outRadius-inRadius)/2;
        startHalfDisk(up, x-d,y,outRadius-d,alpha);
        startHalfDisk(up, x+d,y,outRadius-d,-alpha);
    }
*/

    public void fillArch(boolean up, int x, int y, int inRadius, int outRadius, int alpha)
    {
        if (inRadius>=outRadius)
            throw new IllegalArgumentException("inRadius was >= outRadius.");
        if (x-outRadius<0 || y-outRadius<0 || x+outRadius>=w || y+outRadius>=h) return;
        halfDisk(up, x,y,outRadius,alpha);
        halfDisk(up, x,y,inRadius,-alpha);
    }

    private void startHalfEllipse(boolean up, int x1, int x2, int y, int h, int alpha)
    {
        int radius=(x2-x1)/2;
        int mid=(x1+x2)/2;
        double r2=radius*radius;
        double q2=r2/(h*h);
        int start=up ? -h : 0;
        int end=up ? 0 : h;
        for (int cy=start; cy<=end; cy++)
        {
            int cx=(int)Math.sqrt(r2-cy*cy*q2);
            int b1=(cy+y)*w+mid+cx;
            int b2=(cy+y)*w+mid-cx;
            pix[b1]+=alpha;
            pix[b2]+=alpha;
        }
    }

    private void startHalfDisk(boolean up, int x, int y, int radius, int alpha)
    {
        int r2=radius*radius;
        int start=up ? -radius : 0;
        int end=up ? 0 : radius;
        for (int cy=start; cy<=end; cy++)
        {
            int cx=(int)Math.sqrt(r2-cy*cy);
            int b1=(cy+y)*w+x+cx;
            int b2=(cy+y)*w+x-cx;
            pix[b1]+=alpha;
            pix[b2]+=alpha;
        }
    }

    private void halfDisk(boolean up, int x, int y, int radius, int alpha)
    {
        int r2=radius*radius;
        int start=up ? -radius : 0;
        int end=up ? 0 : radius;
        for (int cy=start; cy<=end; cy++)
        {
            int cx=(int)Math.sqrt(r2-cy*cy);
            int b1=(cy+y)*w+x+cx;
            int b2=(cy+y)*w+x-cx;
            pix[b1]-=alpha;
            pix[b2]+=alpha;
        }
    }

    public void render(Graphics g, int x, int y)
    {
        Image image=makeImage();
        g.drawImage(image,x,y,null);
        image.flush();
    }

    public Image makeImage()
    {
        fill();
        Image image= Toolkit.getDefaultToolkit().createImage(new MemoryImageSource(w, h, model, imagePix, 0, w));
        return image;
    }

    public void clear()
    {
        for (int i=pix.length-1; i>=0; i--)
            pix[i]=0;
    }

    public void time()
    {
        int n=100;
        int i;

 /*
        // clear.
        long start=System.currentTimeMillis();
        for (i=0; i<n; i++)
            clear();
        long time=(System.currentTimeMillis()-start)/n;
        System.out.println("clear time: "+time);
        */


        // fill.
        long start=System.currentTimeMillis();
        for (i=0; i<n; i++)
            fill();
        long fillTime=(System.currentTimeMillis()-start)/n;
        System.out.println("fill time: "+fillTime);


    }


    public void fill()
    {
        int n=0;
        short[] pix=this.pix;            // it is said that local variables
        byte[] imagePix=this.imagePix; // are fastest...
        for (int y=0; y<h; y++)
        {
            int a=0;
            for (int x=w; x>0; x--)
            {
                a+=pix[n];
                byte b;
                if (a<0) b=0;
                else if (a>255) b=-1;
                else b=(byte)a;
                imagePix[n]=b;
                n++;
            }
        }

    }

}