package x.cluster;

import java.awt.*;
import java.util.*;
import java.util.List;

public class ClusterNode
{
	private ClusterNode parent;
	private ClusterNode[] children;
	private int childCount=0;
	private int dataIndex=-1;
	private Rectangle bounds;
	private int numDescendants;
	private int numMarked;
	private int id;
	private static int idCount;
	private boolean significant;
	private float value;
	private boolean isError;
	
	public ClusterNode()
	{
	    id=idCount++;
	    children=new ClusterNode[2];
	}
	
	public void setError(boolean isError)
	{
		this.isError=isError;
	}
	
	public boolean getError()
	{
		return isError;
	}
	
	public ClusterNode(int dataIndex)
	{
	    this();
	    this.dataIndex=dataIndex;
	}
	
	public ClusterNode(int dataIndex, float value)
	{
	    this();
	    this.dataIndex=dataIndex;
	    this.value=value;
	}
	
	public ClusterNode[] getLeaves()
	{
		ArrayList list=new ArrayList();
		addLeaves(list);
		return (ClusterNode[])list.toArray(new ClusterNode[0]);
	}
	
	void addLeaves(List list)
	{
		if (!hasChildren())
			list.add(this);
		else
		{
			getChild(0).addLeaves(list);
			getChild(1).addLeaves(list);
		}
	}
	
    
    public int[] makeDataIndexList()
    {
        int n=countLeaves();
        int[] order=new int[n];
        makeOrder(order, 0);
        return order;
    }
    
    public ClusterNode[] makeIndexedList()
    {
        int n=countLeaves();
        ClusterNode[] list=new ClusterNode[n];
        placeInArrayByIndex(list);
        return list;
    }
    
    private int makeOrder(int[] order, int index)
    {
        if (!hasChildren())
        {
            order[index]=getDataIndex();
            return index+1;
        }
        
        index=getChild(0).makeOrder(order, index);
        return getChild(1).makeOrder(order, index);
    }
    
    private void placeInArrayByIndex(ClusterNode[] list)
    {
        if (!hasChildren())
        {
            list[getDataIndex()]=this;
            return;
        }
        
        getChild(0).placeInArrayByIndex(list);
        getChild(1).placeInArrayByIndex(list);
    }
    
	
	
	public boolean hasData()
	{
	    return dataIndex>=0;
	}
	
	public boolean getSignificant()
	{
	    return significant;
	}
	
	public void setSignificant(boolean significant)
	{
	    this.significant=significant;
	}
	
	public void resetSignificance()
	{
	    setSignificant(false);
	    for (int i=getChildCount()-1; i>=0; i--)
	        getChild(i).resetSignificance();
	}
	
	public void setChildren(ClusterNode c1, ClusterNode c2)
	{
	    childCount=2;
	    int m=c1.getNumDescendants()+1;
	    int n=c2.getNumDescendants()+1;
	    value=(m*c1.value+n*c2.value)/(m+n);
	    addChild(c1,0);
	    addChild(c2,1);
	}
	
	public int getChildCount()
	{
	    return childCount;
	}
	
    private void addChild(ClusterNode child, int index) 
    {
	    child.parent = this;
	    children[index]=child;
	    numDescendants+=(1+child.getNumDescendants());
    }
    
    public void setNumMarked(int numMarked)
    {
        this.numMarked=numMarked;
    }
    
    public int getNumMarked()
    {
        return numMarked;
    }
    
    public Rectangle findBoundsFromMarked()
    {
        int m=getNumMarked();
        if (m==0) return null;
        int n=getChildCount();
        if (n==0)
            return bounds;
            
        Rectangle r=null;
            
        for (int i=0; i<n; i++)
        {
            Rectangle s=getChild(i).findBoundsFromMarked();
            if (s==null) 
                continue;

            if (r==null) 
                r=new Rectangle(s.x, s.y, s.width, s.height);
            else 
                r=r.union(s);
        }
        bounds=r;
        //System.out.println(r);
        return r;
    }
    
    public Rectangle findBoundsFromChildren()
    {
        int n=getChildCount();
        if (n==0) return bounds;
        Rectangle r=null;
        boolean okChild=false;
        for (int i=0; i<n; i++)
        {
            Rectangle s=getChild(i).findBoundsFromChildren();
            if (s==null) 
                continue;
            okChild=true;
            if (r==null) 
                r=new Rectangle(s.x, s.y, s.width, s.height);
            else 
                r=r.union(s);
        }
        bounds=r;
        //System.out.println(r);
        return r;
    }
    
    public int countNumMarked()
    {
        int n=getChildCount();
        if (n==0) return numMarked;
        int s=0;
        for (int i=0; i<n; i++)
            s+=getChild(i).countNumMarked();
        numMarked=s;
        return s;
    }
    
    public int countLeaves()
    {
        if (!hasChildren()) return 1;
        return children[0].countLeaves()+children[1].countLeaves();
    }
    
    public int getNumDescendants()
    {
        return numDescendants;
    }
    
    public void includeInBounds(int x, int y)
    {
        if (bounds==null)
            bounds=new Rectangle(x,y,0,0);
        bounds.add(x,y);
    }
    
    public Rectangle getBounds()
    {
        return bounds;
    }

    public void setDataIndex(int dataIndex)
    {
        this.dataIndex=dataIndex;
    }
    
    public int getDataIndex()
    {
        return dataIndex;
    }

    public boolean childInTree(ClusterNode m) 
    {
        return m==this || m!=null && childInTree(m.parent);
    }


    public ClusterNode getChild(int index) 
    {
	    return children[index];
    }

    public boolean hasChildren() 
    {
	    return children[0]!=null;
    }
    
    public ClusterNode getRoot()
    {
        return parent==null ? this : parent.getRoot();
    }
    
    public String toString()
    {
        return "ClusterNode #"+id+"; data="+dataIndex;
    }
    
    public int getDepth()
    {
        return parent==null ? 0 : 1+parent.getDepth();
    }
    
    public float getValue()
    {
        return value;
    }
    
    public int getID()
    {
        return id;
    }
    
    public int commonAncestorDepth(ClusterNode c)
    {
        if (c==null) throw new NullPointerException("Can't common ancestor to null node!");
        if (childInTree(c))
            return getDepth();
        if (parent==null) throw new IllegalArgumentException("not in same tree!");
        return parent.commonAncestorDepth(c);
    }
    
    public int distance(ClusterNode c)
    {
        if (c==null) throw new NullPointerException("Can't find distance to null node!");
        if (childInTree(c))
            return Math.abs(getDepth()-c.getDepth());
        if (parent==null) throw new IllegalArgumentException("not in same tree!");
        return 1+parent.distance(c);
    }
}
