package x.data;

import x.util.*;
import java.io.*;
import java.util.*;

public class DataArray
{
    private DataRow[] data;
    private MetaData metaData;
    private float nullDataValue=0;
    private String[] fieldNames;
    private DataField[] fields;
    private boolean padShortLines=true;
    
    public DataArray(MetaData metaData)
    {
        this.metaData=metaData;
    }
    
    public DataArray(String metaFileName, boolean verbose, int maxLines) throws IOException, DataParseException
    {
        File file=new File(metaFileName);
        metaData=new MetaData(file);
        readFromFile(verbose, maxLines);
    }
    
    public DataArray(String metaFileName, boolean verbose) throws IOException, DataParseException
    {
        this(metaFileName, verbose, Integer.MAX_VALUE);
    }
    
    public DataArray(String metaFileName) throws IOException, DataParseException
    {
        this(metaFileName, false);
    }
    
    public MetaData getMetaData()
    {
        return metaData;
    }
    
    public DataField[] getFields()
    {
        return fields;
    }
    
    public int size()
    {
        return data==null ? 0 : data.length;
    }
    
    public void setNullDataValue(float nullDataValue)
    {
        this.nullDataValue=nullDataValue;
    }
    
    public DataRow getRow(int i)
    {
        return data[i];
    }
    
    public float num(int row, int column)
    {
        return data[row].getNumbers()[column];
    }
    
    public DataRow[] getData()
    {
        return data;
    }
    
    public float[][] copyNumbers()
    {
        int n=data.length;
        float[][] f=new float[n][];
        for (int i=0; i<n; i++)
        {
            float[] original=data[i].getNumbers();
            int m=original.length;
            f[i]=new float[m];
            System.arraycopy(original, 0, f[i], 0, m);
        }
        return f;
    }
    
    public String[][] copyText()
    {
        int n=data.length;
        String[][] s=new String[n][];
        for (int i=0; i<n; i++)
        {
            String[] original=data[i].getText();
            int m=original.length;
            s[i]=new String[m];
            System.arraycopy(original, 0, s[i], 0, m);
        }
        return s;
    }
    
    public void readFromFile() throws IOException, DataParseException
    {
        readFromFile(false);
    }
    
    public void readFromFile(boolean verbose) throws IOException, DataParseException
    {
        readFromFile(verbose, Integer.MAX_VALUE);
    }
    
    public void readFromFile(boolean verbose, int maxLines) throws IOException, DataParseException
    {
        // Open file.
        
        File file=new File(metaData.getFileName());
        if (verbose)
            System.err.println("Reading data from file: "+file);
        InputStream in=new FileInputStream(file);
        BufferedReader r=new BufferedReader(new InputStreamReader(in));
        
        try
        {
            read(verbose, maxLines, r);
        }
        finally
        {
            r.close();
            in.close();
        }
    }
    
    public void read(boolean verbose, BufferedReader r) throws IOException, DataParseException
    {
        read(verbose, Integer.MAX_VALUE, r);
    }
    
    public void read(boolean verbose, int maxLines, BufferedReader r) throws IOException, DataParseException
    {
        long startTime=System.currentTimeMillis();
        
        // Read the header row and
        // create an array with the type of each line.
        String line=r.readLine();
        FlexibleTokenizer t=new FlexibleTokenizer(line, "\t");
        int n=t.countTokens();
        
        fields=new DataField[n];
        int[] fieldType=new int[n];
        fieldNames=new String[n];
        int numberFieldCount=0;
        int textFieldCount=metaData.getTextFieldCount();
        int idFieldIndex=-1;

        // Figure our which are the number fields, and how many;
        // also, find the index of the id field.
        int i=0;
        int realCount=0;
        int textSoFar=0;
        while (t.hasMoreTokens())
        {
            String fieldName=t.nextToken();
            fieldNames[i]=fieldName;
            fieldType[i]=metaData.getFieldType(fieldName);
            if (fieldType[i]==metaData.NUMBER)
                numberFieldCount++;
            if (fieldType[i]==metaData.TEXT)
                textSoFar++;
            if (metaData.isIDField(fieldName))
                idFieldIndex=i;
            if (fieldType[i]!=metaData.IGNORE)
            {
                fields[realCount]=new DataField(fieldType[i], 
                    fieldType[i]==metaData.TEXT ? textSoFar-1 : numberFieldCount-1, fieldName);
                realCount++;
            }
            i++;
        }
        
        if (realCount!=fields.length)
        {
            DataField[] shortFields=new DataField[realCount];
            System.arraycopy(fields,0,shortFields,0,realCount);
            fields=shortFields;
        }
        
        if (verbose)
        {
            System.err.println("Total fields: "+n);
            System.err.println("Data fields: "+numberFieldCount);
            System.err.println("Text fields: "+textFieldCount);
            System.err.println("ID field index: "+idFieldIndex);
        }
        
        // If an ID field was specified but not found, complain.
        if (metaData.getIDField()!=null && idFieldIndex==-1)
            throw new DataParseException("This ID field: "+metaData.getIDField()+
                " wasn't found on this header line: "+line);
        
        // Read the data.
        Vector rowList=new Vector();
        int linesRead=0;
        
        // Set up list of items to ignore.
        Vector starts=metaData.getIgnoredRowStarts();
        int q=starts.size();
        String[] ignoreStart=new String[q];
        for (i=0; i<q; i++)
            ignoreStart[i]=(String)starts.elementAt(i);
        
        System.out.print("Reading data");
        while (null!=(line=r.readLine()))
        {
            // If we've been told to ignore this line,
            // then ignore it!
            boolean ignoring=false;
            for (i=0; i<q; i++)
                if (line.startsWith(ignoreStart[i]))
                {
                    ignoring=true;
                    break;
                }
            if (ignoring)
            {
                if (verbose)
                    System.err.println("\nIgnoring line starting with: "+ignoreStart[i]);
                continue;
            }
            
            
            float[] numbers=new float[numberFieldCount];
            String[] text=new String[textFieldCount];
            
            try
            {
                t=new FlexibleTokenizer(line, "\t");
                if (t.countTokens()!=fieldType.length)
                {
                	String warning="Line has different number of fields than in header row:\n "+
					t.countTokens()+" instead of "+fieldType.length+":\n"+line;
                    if (padShortLines)
                    	System.err.println("Warning: "+warning);
                    else
                    	throw new DataParseException(warning);
                }
                
                DataRow row=new DataRow();
                int fieldNum=0;
                int numberFieldNum=0;
                int textFieldNum=0;
                while (t.hasMoreTokens())
                {
                    String s=t.nextToken();
                    
                    // Set ID.
                    if (fieldNum==idFieldIndex)
                        row.setID(s);
                    
                    // Read field if text or numbers.
                    //System.out.println(fieldType[fieldNum]+" --> "+s);
                    switch (fieldType[fieldNum])
                    {
                        case 0://metaData.NUMBER:
                            numbers[numberFieldNum++]=parseFloat(s);
                            break;
                        case 2://metaData.TEXT:
                            text[textFieldNum++]=s;
                            break;
                    }
                    
                    fieldNum++;
                }
                rowList.addElement(row);
                if (numberFieldNum!=numberFieldCount && !padShortLines)
                    throw new DataParseException("Wrong number of data fields ("+
                        numberFieldNum+", not "+numberFieldCount+") in "+line);
                row.setNumbers(numbers);
                if (textFieldNum!=textFieldCount && !padShortLines)
                    throw new DataParseException("Wrong number of text fields ("+
                        textFieldNum+", not "+textFieldCount+") in "+line);
                row.setText(text);
            }
            catch (Exception e)
            {
                if (verbose)
                {
                    System.err.println("\n\nException on line "+(2+linesRead));
                    System.err.println(line);
                    e.printStackTrace(System.err);
                }
                throw new DataParseException(e.getMessage());
            }
            
            if (++linesRead>maxLines) break;
            if (verbose && linesRead%500==0)
                System.out.print(".");
        }
        
        n=rowList.size();
        data=new DataRow[n];
        rowList.copyInto(data);
        
        if (verbose)
        {
            System.err.println(" ");
            System.err.println("rows read: "+n);
            System.err.println("time elapsed: "+(System.currentTimeMillis()-startTime));
        }
    }
    
    float parseFloat(String s)
    {
        try
        {
            return Float.valueOf(s).floatValue(); // preserve 1.1 compatibility.
        }
        catch (NumberFormatException e)
        {
            return 0;
        }
    }
    
}