package x.path;

import java.awt.*;
import java.applet.*;

/**

This class creates a space-filling curve with good clustering properties.

The code is adapted from the applet demo on this page:
http://www-fs.informatik.uni-tuebingen.de/~reinhard/hcurve.html
which was written by Klaus Reinhardt
to illustrate the ideas contained in the paper:
    R. Niedermeier , K. Reinhardt and P. Sanders . 
    Towards optimal locality in mesh-indexings. 
    Proceedings of the 11th International Symposium on 
    Fundamentals of Computation Theory, number 1279 in Lecture Notes in Computer Science, 
    pages 364--375, Krakow, Poland, September 1997. Springer.
    
*/

// The HCurve class is essentially a wrapper for Klaus Reinhardt's
// algorithm, which is referenced only in the private "make" method.
    
public class HCurve implements PathLayout
{
    private int width;
    private int height;
    private int length;
    private Point[] path;
    private int[][] order;
    
    public HCurve(int length)
    {
        int n=(int)Math.ceil(Math.sqrt(length));
        int m=(int)Math.floor(Math.sqrt(length));
        if (n==m)
            make(length, n, n);
        else if (n*m>=length)
            make(length, n, m);
        else
            make(length, n, n);
    }
    
    public int order(int x, int y)
    {
        if (x>=0 && y>=0 && x<width && y<height)
            return order[x][y];
        return -1;
    }
    
    public boolean contains(int x, int y)
    {
        int n=order(x,y);
        return n>=0 && n<length;
    }
    
    public int x(int n)
    {
        return path[n].x;
    }
    
    public int y(int n)
    {
        return path[n].y;
    }
    
    public Point getPoint(int n)
    {
        return path[n];
    }
    
    public Dimension getSize()
    {
        return new Dimension(width, height);
    }
    
    public int getLength()
    {
        return length;
    }
   
    private void make(int length, int w, int h)
    {
        width=w;
        height=h;
        this.length=length;
        
        HCurveSegment s= new HCurveMaker(w,h).felds(w,h);
        path=new Point[length];
        order=new int[w][h];
        int x,y;
        for (x=0; x<w; x++)
            for (y=0; y<h; y++)
                order[x][y]=-1;
                
        for (int i=0; i<length; i++)
        {
            x=s.x(i);
            y=s.y(i);
            order[x][y]=i;
            path[i]=new Point(x,y);
        }
            
    }
}


// The following code was created
// by refactoring Klaus Reinhardt's applet source into two classes.
// Note that I kept only the code for doing rectangles of
// arbitrary dimensions, since that is all that I care about
// in my applications.


class HCurveSegment
{
    int x,y,g,h;
    HCurveSegment lo, ro, lu, ru, sp;
   
   public int x(int i) {
     if (x == 1) {return 0;} else {
      if (y == 1) {return i;} else {
       if (2*lo.y == y) {
         if (i < lo.g-lo.h) {return lo.x-1-lo.x(i+lo.h);} else {
          if (i < lu.g+lo.g-lo.h) {
           return lu.x-1-lu.x(lu.g+lo.g -lo.h -i-1);
          } else {
           if (i < ru.g+lu.g+lo.g-lo.h) {
            return lu.x+ru.x(i-lu.g-lo.g +lo.h);
           } else {
            if (i < g-lo.h) {
             return lo.x+ro.x(g -lo.h  -i-1);
            } else {
             return lo.x-1-lo.x(i+lo.h-g);
       } }}}} else {
        if (2*lo.x == x) { return sp.y(g-i);
        } else {
         if (i < lo.h) {return lo.x(i);} else {
          if (i < lu.g+lo.h) {
           return lu.x-1-lu.x(lu.g +lo.h -i-1);
          } else {
           if (i < lu.g+ru.g+lo.h) {
            return lu.x+ru.x(i-lu.g-lo.h);
           } else {
            if (i < g-lo.g+lo.h) {
             return lo.x+ro.x(g -lo.g+lo.h  -i-1);
            } else {
             return lo.x(i+lo.g-g);
     }}}}}}}}
   }
   public int y(int i) {
     if (x == 1) {return i;} else {
      if (y == 1) {return 0;} else {
       if (2*lo.y == y) {
         if (i < lo.g-lo.h) {return lo.y-1-lo.y(i+lo.h);} else {
          if (i < lu.g+lo.g-lo.h) {
           return lo.y+lu.y(lu.g+lo.g -lo.h -i-1);
          } else {
           if (i < ru.g+lu.g+lo.g-lo.h) {
            return ro.y+ru.y(i-lu.g-lo.g +lo.h);
           } else {
            if (i < g-lo.h) {
             return ro.y-1-ro.y(g -lo.h  -i-1);
            } else {
             return lo.y-1-lo.y(i+lo.h-g);
       } }}}} else {
        if (2*lo.x == x) { return sp.x(g-i);
        } else {
         if (i < lo.h) {return lo.y(i);} else {
          if (i < lu.g+lo.h) {
           return lo.y+lu.y(lu.g +lo.h -i-1);
          } else {
           if (i < lu.g+ru.g+lo.h) {
            return ro.y+ru.y(i-lu.g-lo.h);
           } else {
            if (i < g-lo.g+lo.h) {
             return ro.y-1-ro.y(g -lo.g+lo.h  -i-1);
            } else {
             return lo.y(i+lo.g-g);
     }}}}}}}}
   }
}

// There is a separate object to construct a curve segment,
// because it makes it easier to implement a cache
// so we don't keep recalculating small curves. 

class HCurveMaker
{
    HCurveSegment fel[];    // These instance variables allow us to
    HCurveSegment feld[][]; // cache calculations.
        
    HCurveMaker(int w, int h)
    {
        int m=Math.max(w,h)+1;
        fel = new HCurveSegment[m];
        feld = new HCurveSegment[m][m];
    }
        
    HCurveSegment fels(int xm) 
    {
        HCurveSegment r;
        if (fel[xm]!=null) {return fel[xm];} else 
        {
            r = new HCurveSegment(); fel[xm] = r; 
            r.g=xm; r.x=xm; r.y=1; r.h=1;
            r.lo = felds(0,0); r.ro = felds(0,0);
            r.lu = felds(0,0); r.ru = felds(0,0);
            return r;
        }   
    }
        
    HCurveSegment felds(int xm, int ym) 
    {
        HCurveSegment r;
        int xn, yn;

        if (feld[xm][ym]!=null) 
        {
            return feld[xm][ym];
        } else {
        r = new HCurveSegment(); 
        feld[xm][ym] = r; 
        r.x=xm; r.y=ym; xn=xm/2; yn=ym/2; r.g=xm*ym;
        if (xn==(xn/2)*2) {xn= xm -xn;};
        if (yn==(yn/2)*2) {yn= ym -yn;};
        if ((xm-xn)*(ym-yn)==0) {r.h=xm*ym-1;
            if (xm*ym==0) {r.h+=1;};
            r.lo = felds(0,0); r.ro = felds(0,0);
            r.lu = felds(0,0); r.ru = felds(0,0);
        } else {
            r.lo = felds(xn,yn); r.ro = felds(xm-xn,yn);
            r.lu = felds(xn,ym-yn); r.ru = felds(xm-xn,ym-yn);
            if (xm == 3) {
            if (2*yn != ym) {
            r.lo = fels(3); r.ro = felds(0,1);
            r.lu = felds(0,ym-1); r.ru = felds(3,ym-1);
            } else {
            if (2*(yn/2) != yn) {
            r.lo = felds(3,yn); r.ro = felds(0,yn);
            r.lu = felds(0,yn); r.ru = felds(3,yn);
            } else {
            r.lo = felds(3,yn); r.ro = felds(0,yn);
            r.lu = felds(3,yn); r.ru = felds(0,yn);
            }}} else {
            if ((xm == 2) && (2*yn == ym)) {
            r.lo = felds(2,yn); r.ro = felds(0,yn);
            r.lu = felds(0,ym-yn); r.ru = felds(2,ym-yn);
            } else {
            if ((ym == 3) && (2*xn != xm)) {
            r.lo = fels(xm); r.ro = felds(0,1);
            r.lu = felds(0,ym-1); r.ru = felds(xm,ym-1);
            }}}
            if (2*r.lo.y == ym) {r.h = r.lo.g-r.lo.h;} else {r.h = r.lo.h;};
            r.h=r.h+r.lu.g+r.ru.h;
            if ((xm == 3) && (2*(yn/2) == yn) && (2*yn == ym)) {
            r.h = r.h - r.lu.h +1;
        }}
        if ((2*r.lo.x == xm) && (2*r.lo.y != ym)) {
            r.sp = felds(ym,xm);
            r.h = r.sp.g - r.sp.h;
        }
        return r;
    }   }
        
}

